/******************************************************************************

   This file is a part of the

   ICE-POISSON TOOLBOX: Total Variation Restoration of Images Corrupted by
   Poisson Noise with Iterated Conditional Expectations

   Copyright (C) 2018 Rémy Abergel (Remy.Abergel AT parisdescartes.fr), Cécile
   Louchet (Cecile.Louchet AT univ-orleans.fr), Lionel Moisan (Lionel.Moisan AT
   parisdescartes.fr), Tieyong Zeng (zeng AT hkbu.edu.hk).

   This software provides a C implementation of the image denoising algorithm
   described in the following research papers

   [1] R. Abergel, C. Louchet, L. Moisan and T. Zeng, ``Total Variation
   Restoration of Images Corrupted by Poisson Noise with Iterated Conditional
   Expectations''.  Proceedings of the 5th International Conference on Scale
   Space and Variational Methods in Computer Vision (SSVM), pp. 178--190, 2015,
   Springer, DOI: 10.1007/978-3-319-18461-6_15.

   [2] R. Abergel and L. Moisan, ``Fast and accurate evaluation of a generalized
   incomplete gamma function'', Preprint MAP5 2016-14,
   https://hal.archives-ouvertes.fr/hal-01329669.

   This program is free software: you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free Software
   Foundation, either version 3 of the License, or (at your option) any later
   version.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along with
   this program. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

/* compilation (with gcc): gcc -O3 tiffaxpb.c tiffread.c tiffwrite.c -lm -ltiff -o tiffaxpb */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#define FREE_ALL() {					\
    if(NULL != u) free(u);				\
    if(NULL != u0) free(u0);				\
  }
#define ASSERT_ALLOC(cmd) if(NULL == (cmd)) { printf("Not enough memory\n"); FREE_ALL(); exit(EXIT_FAILURE); }

extern int read_tiff_into_double(double**,int*,int*,char*,char); /* see source file 'tiffread.c' */
extern int save_tiff(char*,double*,int,int,unsigned int16,char,char,char); /* see source file 'tiffwrite.c' */
extern void deltagammainc(double*,double*,char*,double,double,double,double); /* see source file 'kernel.c' */

/* display usage */
static void display_usage()
{
  printf("\nGain/offset correction to a TIFF image: out = a*in+b.\n\n");
  printf("Usage: tiffaxpb [-a a] [-b b] [-f] [-v] in out\n\n");
  printf("   -a a : (default = 1.) slope of the affine contrast change\n");
  printf("   -b b : (default = 0.) offset of the affine contrast change\n");
  printf("   -f   : save output as a tiff-float instead of unsigned char tiff\n");
  printf("   -v   : light verbose mode (display progression)\n");
  printf("   in   : input image (TIFF)\n");
  printf("   out  : output image (TIFF)\n\n");
}

int main(int argc, char **argv)
{
  char *filenamein = NULL,*filenameout = NULL,*avalue = NULL,*bvalue = NULL,vflag,fflag;
  double *u0 = NULL,*u = NULL,a,b;
  int i,err,nx,ny,adr;
  vflag = fflag = 0;

  /******************************************/
  /*            arguments parser            */
  /******************************************/

  /* main parser */
  for(i=1;i<argc;i++) {
    if (strcmp(argv[i],"-v") == 0) vflag = 1;
    else if (strcmp(argv[i],"-f") == 0) fflag = 1;
    else if (strcmp(argv[i],"-a") == 0) {
      if(i==argc-1) { display_usage(); printf("Error: option -a requires an argument\n\n"); return EXIT_FAILURE; }
      else {
	avalue = argv[i+1]; err = sscanf(avalue,"%lf",&a); i++;
	if(err != 1) { display_usage(); printf("Error: could not retrieve properly the argument of option -a\n\n"); return EXIT_FAILURE; }
      }
    }
    else if (strcmp(argv[i],"-b") == 0) {
      if(i==argc-1) { display_usage(); printf("Error: option -b requires an argument\n\n"); return EXIT_FAILURE; }
      else {
	bvalue = argv[i+1]; err = sscanf(bvalue,"%lf",&b); i++;
	if(err != 1) { display_usage(); printf("Error: could not retrieve properly the argument of option -b\n\n"); return EXIT_FAILURE; }
      }
    }
    else if (NULL == filenamein) filenamein = argv[i];
    else if (NULL == filenameout) filenameout = argv[i];
    else { display_usage(); printf("Error: could not retrieve properly input arguments.\n\n"); return EXIT_FAILURE; }
  }

  /* set default values */
  if(NULL == avalue) a = 1.;
  if(NULL == bvalue) b = 0.;

  /* check consistency */
  if ((NULL == filenamein) || (NULL == filenameout)) {
    display_usage();
    if (NULL == filenamein) printf("Error: input 'in' is missing\n\n");
    else if(NULL == filenameout) printf("Error: output 'out' is missing\n\n");
    return EXIT_FAILURE;
  }

  /* deal with verbose mode */
  if(vflag) {
    printf("\nInput parameters\n");
    printf("================\n");
    printf("affine slope: a = %g\n",a);
    printf("affine offset: b = %g\n\n",b);
  }

  /************************************************************/
  /* Open TIFF Image & get its graylevels in double precision */
  /************************************************************/
  if(vflag) {
    printf("Load input image and allocate memory\n");
    printf("====================================\n");
  }

  if(EXIT_FAILURE == read_tiff_into_double(&u0,&nx,&ny,filenamein,vflag)) {
    if(vflag) printf("Free memory and exit the program.\n\n");
    else printf("An error occured (try verbose mode using option -v to get more informations)\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  /*********************/
  /* Memory allocation */
  /*********************/
  ASSERT_ALLOC(u = (double*) malloc(nx*ny * sizeof(double)));

  /***********************************/
  /* KERNEL: process the input image */
  /***********************************/
  if(vflag) {
    printf("Apply contrast change out = a*in + b\n");
    printf("====================================\n");
  }
  for(adr=0; adr<nx*ny; adr++) u[adr] = a*u0[adr] + b;
  if(vflag) printf("done.\n\n");

  /***********************************/
  /*          END OF KERNEL          */
  /***********************************/

  /********************************/
  /* Write output as a TIFF Image */
  /********************************/
  if(vflag) {
    printf("Save output image in TIFF format\n");
    printf("================================\n");
  }

  if (EXIT_FAILURE == save_tiff(filenameout,u,nx,ny,8,fflag,1,vflag)) {
    if (vflag) printf("free memory and exit the program.\n\n");
    else printf("An error occured (try verbose mode using option -v to get more informations)\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  FREE_ALL();

  return EXIT_SUCCESS;
}
